from ase.geometry.dimensionality.interval_analysis import (
    analyze_kintervals as analyze_dimensionality,
)
from ase.geometry.dimensionality.isolation import isolate_components

__all__ = ['analyze_dimensionality', 'isolate_components']
