raise ImportError(
    'The ase.ga.bulk_startgenerator module has been deprecated. '
    'The same functionality is now provided by the '
    'ase.ga.startgenerator module. Please consult its documentation '
    'to verify how to e.g. initialize a StartGenerator object.')
